#!/usr/bin/env python3
"""
HTTP Protocol Fuzzer
Parameters:
    - IP: Target IP address
    - port: Target port number
Additional parameters in comments:
    - wordlist: Path to wordlist file
    - threads: Number of concurrent threads
Warning: Use for security testing only
"""
import requests
import argparse
import threading
import queue
import time
from datetime import datetime
import random
import string
import logging
from pathlib import Path

class HTTPFuzzer:
    def __init__(self, target_ip, port, threads=10, timeout=5):
        self.target_ip = target_ip
        self.port = port
        self.threads = threads
        self.timeout = timeout
        self.url_queue = queue.Queue()
        self.results = queue.Queue()
        self.stop_event = threading.Event()
        self.setup_logging()
        
        # Common HTTP methods
        self.http_methods = ['GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'OPTIONS', 'TRACE']
        
        # Common parameters to fuzz
        self.common_params = ['id', 'user', 'name', 'file', 'dir', 'path', 'cmd']
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("fuzzing_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "http_fuzzer.log"),
                logging.StreamHandler()
            ]
        )
        
    def generate_random_string(self, length=10):
        """Generate random string for fuzzing"""
        return ''.join(random.choices(string.ascii_letters + string.digits, k=length))
        
    def generate_payload(self):
        """Generate fuzzing payload"""
        payload_types = [
            lambda: self.generate_random_string(),
            lambda: "A" * random.randint(100, 1000),
            lambda: f"'{self.generate_random_string()}'",
            lambda: f"{random.randint(-1000000, 1000000)}",
            lambda: f"../../{self.generate_random_string()}",
            lambda: f"<script>{self.generate_random_string()}</script>",
            lambda: f"$(cat /etc/passwd)",
            lambda: f"%00{self.generate_random_string()}",
            lambda: f"{self.generate_random_string()}/*",
            lambda: f"{{self.generate_random_string()}}"
        ]
        
        return random.choice(payload_types)()
        
    def generate_headers(self):
        """Generate random HTTP headers"""
        headers = {
            'User-Agent': f'FuzzBot/{self.generate_random_string(5)}',
            'Accept': '*/*',
            'Connection': 'keep-alive'
        }
        
        # Add random custom headers
        for _ in range(random.randint(0, 5)):
            headers[f'X-Fuzz-{self.generate_random_string(5)}'] = self.generate_random_string()
            
        return headers
        
    def generate_urls(self):
        """Generate URLs for fuzzing"""
        common_paths = ['/', '/api', '/admin', '/login', '/upload', '/download']
        
        for base_path in common_paths:
            # Basic path fuzzing
            self.url_queue.put((base_path, 'GET', None))
            
            # Path traversal attempts
            traversal_path = f"{base_path}/../../{self.generate_random_string()}"
            self.url_queue.put((traversal_path, 'GET', None))
            
            # Parameter fuzzing
            for param in self.common_params:
                path = f"{base_path}?{param}={self.generate_payload()}"
                self.url_queue.put((path, 'GET', None))
                
                # POST data
                data = {param: self.generate_payload()}
                self.url_queue.put((base_path, 'POST', data))
                
    def fuzzer_thread(self):
        """Individual fuzzer thread"""
        session = requests.Session()
        
        while not self.stop_event.is_set():
            try:
                path, method, data = self.url_queue.get_nowait()
            except queue.Empty:
                break
                
            url = f"http://{self.target_ip}:{self.port}{path}"
            headers = self.generate_headers()
            
            try:
                start_time = time.time()
                response = session.request(
                    method=method,
                    url=url,
                    headers=headers,
                    data=data,
                    timeout=self.timeout,
                    verify=False  # Disable SSL verification for testing
                )
                duration = time.time() - start_time
                
                result = {
                    'url': url,
                    'method': method,
                    'status_code': response.status_code,
                    'response_time': duration,
                    'response_length': len(response.content),
                    'timestamp': datetime.now().isoformat()
                }
                
                # Log interesting findings
                if response.status_code not in [404, 403, 400]:
                    logging.info(f"Interesting response found: {result}")
                    self.results.put(result)
                    
            except requests.exceptions.RequestException as e:
                logging.error(f"Error fuzzing {url}: {str(e)}")
                
            self.url_queue.task_done()
            
    def start_fuzzing(self):
        """Start fuzzing process"""
        logging.info(f"Starting HTTP fuzzing against {self.target_ip}:{self.port}")
        logging.info(f"Using {self.threads} threads")
        
        # Generate URLs for fuzzing
        self.generate_urls()
        total_requests = self.url_queue.qsize()
        logging.info(f"Generated {total_requests} test cases")
        
        # Start fuzzer threads
        threads = []
        for _ in range(min(self.threads, total_requests)):
            thread = threading.Thread(target=self.fuzzer_thread)
            thread.start()
            threads.append(thread)
            
        try:
            # Monitor progress
            start_time = time.time()
            while any(t.is_alive() for t in threads):
                remaining = self.url_queue.qsize()
                completed = total_requests - remaining
                elapsed = time.time() - start_time
                rate = completed / elapsed if elapsed > 0 else 0
                
                print(f"\rProgress: {completed}/{total_requests} "
                      f"({completed/total_requests*100:.1f}%) "
                      f"Rate: {rate:.1f} requests/sec", end="")
                
                time.sleep(0.1)
                
        except KeyboardInterrupt:
            logging.info("\nFuzzing stopped by user")
            self.stop_event.set()
            
        # Wait for threads to finish
        for thread in threads:
            thread.join()
            
        # Save results
        self.save_results(total_requests, time.time() - start_time)
        
    def save_results(self, total_requests, duration):
        """Save fuzzing results to file"""
        interesting_results = []
        while not self.results.empty():
            interesting_results.append(self.results.get())
            
        report = {
            'target': f"{self.target_ip}:{self.port}",
            'timestamp': datetime.now().isoformat(),
            'duration': duration,
            'total_requests': total_requests,
            'interesting_findings': interesting_results
        }
        
        report_file = Path("fuzzing_logs") / f"report_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(report_file, 'w') as f:
            json.dump(report, f, indent=2)
            
        logging.info(f"\nFuzzing complete. Found {len(interesting_results)} interesting responses")
        logging.info(f"Report saved to {report_file}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="HTTP Protocol Fuzzer")
    parser.add_argument("ip", help="Target IP address")
    parser.add_argument("port", type=int, help="Target port number")
    parser.add_argument("--threads", type=int, default=10,
                      help="Number of concurrent threads")
    parser.add_argument("--timeout", type=float, default=5,
                      help="Request timeout in seconds")
    args = parser.parse_args()
    
    fuzzer = HTTPFuzzer(args.ip, args.port, args.threads, args.timeout)
    fuzzer.start_fuzzing()
